/*-------------- Telecommunications & Signal Processing Lab ---------------

Routine:
  AFILE *AFrdWVhead (FILE *fp)

Purpose:
  Get file format information from a RIFF WAVE file

Description:
  This routine reads the header for a RIFF WAVE file.  The header information
  is used to set the file data format information in the audio file pointer
  structure.

  RIFF WAVE file:
   Offset Length Type    Contents
      0     4    char   "RIFF" file identifier
      4     4    int    Chunk length
      8     4    char   "WAVE" file identifier
    ...   ...    ...    ...
      F     4    char   "fmt " chunk identifier
     +4     4    int    Chunk length
     +8     2    int      Audio data type
    +10     2    int      Number of interleaved channels
    +12     4    int      Sample rate
    +16     4    int      Average bytes/sec
    +20     2    int      Block align
    +22     2    int      Data word length (bits)
    +24     2    int      Extra data count
    ...    ...   ...
      C     4    char   "fact chunk identifier
     +4     4    int    Chunk length
     +8     4    int      Number of samples
    ...    ...   ...    ...
      D     4    char   "data" chunk identifier
     +4     4    int    Chunk length
     +8    ...   ...      Audio data

Parameters:
  <-  AFILE *AFrdWVhead
      Audio file pointer for the audio file
   -> FILE *fp
      File pointer for the file

Author / revision:
  P. Kabal  Copyright (C) 2001

  $Revision: 1.69 $  $Date: 2001/04/09 19:35:54 $

-------------------------------------------------------------------------*/

static char rcsid [] = "$Id: AFrdWVhead.c 1.69 2001/04/09 AFsp-v5r2 $";

#include <setjmp.h>
#include <string.h>

#include <libtsp.h>
#include <libtsp/AFdataio.h>
#include <libtsp/AFheader.h>
#include <libtsp/AFmsg.h>
#include <libtsp/AFpar.h>
#include <libtsp/WVpar.h>
#include <libtsp/nucleus.h>

#define ICEILV(n,m)	(((n) + ((m) - 1)) / (m))	/* int n,m >= 0 */
#define NELEM(array)	((sizeof array) / (sizeof array[0]))
#define RNDUPV(n,m)	((m) * ICEILV (n, m))		/* Round up */

#define SAME_CSTR(str,ref) 	(memcmp (str, ref, sizeof (str)) == 0)

#define ALIGN		2	/* Chunks padded out to a multiple of ALIGN */

#define WV_FMT_CKSIZE	16
#define WV_LHMIN	(8 + 4 + 8 + WV_FMT_CKSIZE + 8)

/* setjmp / longjmp environment */
extern jmp_buf AFR_JMPENV;

static void
AF_decFMT (struct WV_CKfmt *CKfmt, struct AF_read *AFr);
static int
AF_rdDISP_text (FILE *fp, int Size, struct AF_info *Hinfo);
static int
AF_rdFMT (FILE *fp, struct WV_CKfmt *CKfmt);
static int
AF_rdFACT (FILE *fp, struct WV_CKfact *CKfact);
static int
AF_rdLIST_INFO (FILE *fp, int Size, struct AF_info *Hinfo);
static void
AF_rdRIFF_WAVE (FILE *fp, struct WV_CKRIFF *CKRIFF);
static void
AF_UnsFormat (int FormatTag);


AFILE *
AFrdWVhead (FILE *fp)

{
  AFILE *AFp;
  int AtData, FoundAFsp;
  long int offs, LRIFF, Fact_Nsamp, Dstart, Dend;
  char Info[AF_MAXINFO];
  struct WV_CKRIFF CKRIFF;
  struct WV_CKpreamb CkHead;
  struct AF_read AFr;

/* Set the long jump environment; on error return a NULL */
  if (setjmp (AFR_JMPENV))
    return NULL;	/* Return from a header read error */

/* Defaults and initial values */
  AFr.DFormat.Format = FD_UNDEF;
  AFr.NData.Nsamp = AF_NSAMP_UNDEF;
  AFr.NData.Ldata = AF_LDATA_UNDEF;
  AFr.Hinfo.Info = Info;
  AFr.Hinfo.N = 0;
  AFr.Hinfo.Nmax = AF_MAXINFO;
  Fact_Nsamp = AF_NSAMP_UNDEF;

/* Check the file magic for a RIFF/WAVE file */
  AF_rdRIFF_WAVE (fp, &CKRIFF);
  offs = 12L;	/* Positioned after RIFF/WAVE preamble */
  LRIFF = CKRIFF.ckSize + 8;

  Dstart = 0L;
  AtData = 0;
  FoundAFsp = 0;
  while (offs < LRIFF) {

    /* Read the chunk preamble */
    offs += RHEAD_S (fp, CkHead.ckID);

    /* fmt chunk */
    if (SAME_CSTR (CkHead.ckID, "fmt ")) {
      offs += AF_rdFMT (fp, &CKRIFF.CKfmt);
      AF_decFMT (&CKRIFF.CKfmt, &AFr);
    }

    /* fact chunk */
    else if (SAME_CSTR (CkHead.ckID, "fact")) {
      offs += AF_rdFACT (fp, &CKRIFF.CKfact);
      Fact_Nsamp = CKRIFF.CKfact.dwSampleLength;
    }

    /* data chunk */
    else if (SAME_CSTR (CkHead.ckID, "data")) {
      offs += RHEAD_V (fp, CKRIFF.CKdata.ckSize, DS_EL);
      AFr.NData.Ldata = CKRIFF.CKdata.ckSize;
      Dstart = offs;
      Dend = offs + CKRIFF.CKdata.ckSize;
      if (RNDUPV (Dend, ALIGN) >= LRIFF || ! FLseekable (fp)) {
	AtData = 1;
	break;
      }
      else {
	offs += RSKIP (fp, RNDUPV (CKRIFF.CKdata.ckSize, ALIGN));
	AtData = 0;
      }
    }

    /* Text chunks */
    else if (SAME_CSTR (CkHead.ckID, "afsp")) {
      offs += RHEAD_V (fp, CkHead.ckSize, DS_EL);
      offs += AFrdHinfo (fp, (int) CkHead.ckSize, &AFr.Hinfo, ALIGN);
      FoundAFsp = 1;
    }
    else if (SAME_CSTR (CkHead.ckID, "DISP") && ! FoundAFsp) {
      offs += RHEAD_V (fp, CkHead.ckSize, DS_EL);
      offs += AF_rdDISP_text (fp, (int) CkHead.ckSize, &AFr.Hinfo);
    }
    else if (SAME_CSTR (CkHead.ckID, "LIST") && ! FoundAFsp) {
      offs += RHEAD_V (fp, CkHead.ckSize, DS_EL);
      offs += AF_rdLIST_INFO (fp, (int) CkHead.ckSize, &AFr.Hinfo);
    }
    /* Miscellaneous chunks */
    else {
      offs += RHEAD_V (fp, CkHead.ckSize, DS_EL);
      offs += RSKIP (fp, RNDUPV (CkHead.ckSize, ALIGN));
    }
  }

  if (AFr.DFormat.Format == FD_UNDEF || AFr.NData.Ldata < 0L) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadHead);
    return NULL;
  }
  if (! AtData && offs != LRIFF)
    UTwarn ("AFrdWVhead - %s", AFM_WV_FixRIFF);

  /* Use the number of samples from the fact chunk only for non-PCM
     files.  Some PCM files use the fact chunk, but it does NOT
     specify the number of samples */
  if (CKRIFF.CKfmt.wFormatTag != WAVE_FORMAT_PCM
      && Fact_Nsamp != AF_NSAMP_UNDEF)
    AFr.NData.Nsamp = Fact_Nsamp;

  /* Position at the start of data */
  if (! AtData) {
    if (AFseek (fp, Dstart, NULL))
      return NULL;
  }

/* Set the parameters for file access */
  AFp = AFsetRead (fp, FT_WAVE, &AFr, AF_FIX_LDATA_HIGH);

  return AFp;
}

/* Check the RIFF/WAVE file preamble:  On error return via longjmp */


static void
AF_rdRIFF_WAVE (FILE *fp, struct WV_CKRIFF *CKRIFF)

{
  long int Lfile, LRIFF;

  RHEAD_S (fp, CKRIFF->ckID);
  if (! SAME_CSTR (CKRIFF->ckID, "RIFF")) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  RHEAD_V (fp, CKRIFF->ckSize, DS_EL);
  LRIFF = CKRIFF->ckSize + 8;
  if (LRIFF < WV_LHMIN) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadRIFF);
    longjmp (AFR_JMPENV, 1);
  }

  if (FLseekable (fp)) {
    Lfile = FLfileSize (fp);
    if (LRIFF > Lfile) {
      CKRIFF->ckSize = Lfile - 8;
      UTwarn ("AFrdWVhead - %s", AFM_WV_FixRIFF);
    }
  }
 
  RHEAD_S (fp, CKRIFF->WAVEID);
  if (! SAME_CSTR (CKRIFF->WAVEID, "WAVE")) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadId);
    longjmp (AFR_JMPENV, 1);
  }

  return;
}

/* Read the format chunk, starting at ckSize */


static int
AF_rdFMT (FILE *fp, struct WV_CKfmt *CKfmt)

{
  int offs;

  offs = RHEAD_V (fp, CKfmt->ckSize, DS_EL);
  if (CKfmt->ckSize < WV_FMT_CKSIZE) {
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadPCM);
    longjmp (AFR_JMPENV, 1);
  }

  offs += RHEAD_V (fp, CKfmt->wFormatTag, DS_EL);
  offs += RHEAD_V (fp, CKfmt->nChannels, DS_EL);
  offs += RHEAD_V (fp, CKfmt->nSamplesPerSec, DS_EL);
  offs += RHEAD_V (fp, CKfmt->nAvgBytesPerSec, DS_EL);
  offs += RHEAD_V (fp, CKfmt->nBlockAlign, DS_EL);
  offs += RHEAD_V (fp, CKfmt->wBitsPerSample, DS_EL);

  /* Skip over any extra data at the end of the fmt chunk */
  if (offs > (int) CKfmt->ckSize + 4)
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadPCM);
  else
    offs += RSKIP (fp, RNDUPV (CKfmt->ckSize + 4, ALIGN) - offs);

  return offs;
}

/* Decode the data format: On error return via longjmp */


static void
AF_decFMT (struct WV_CKfmt *CKfmt, struct AF_read *AFr)

{
  switch (CKfmt->wFormatTag) {
  case WAVE_FORMAT_PCM:
    if (CKfmt->wBitsPerSample > 8 && CKfmt->wBitsPerSample <= 16) {
      AFr->DFormat.Format = FD_INT16;
      AFr->DFormat.ScaleF = WV_SF_PCM16;
    }
    /* Special case: IEEE float from CoolEdit (flagged as 24-bit integer) */
    else if (CKfmt->wBitsPerSample == 24 && 
	     CKfmt->nAvgBytesPerSec == 4 * CKfmt->nChannels *
                                       CKfmt->nSamplesPerSec) {
      AFr->DFormat.Format = FD_FLOAT32;
      AFr->DFormat.ScaleF = WV_SF_FLOAT32_X;
      CKfmt->wBitsPerSample = 8 * FDL_FLOAT32;
      CKfmt->wFormatTag = WAVE_FORMAT_IEEE_FLOAT;
      if (! UTcheckIEEE ())
	UTwarn ("AFrdAIhead - %s", AFM_WV_NoIEEE);
    }
    else if (CKfmt->wBitsPerSample > 16 && CKfmt->wBitsPerSample <= 24) {
      AFr->DFormat.Format = FD_INT24;
      AFr->DFormat.ScaleF = WV_SF_PCM24;
    }
    else if (CKfmt->wBitsPerSample > 24 && CKfmt->wBitsPerSample <= 32) {
      AFr->DFormat.Format = FD_INT32;
      AFr->DFormat.ScaleF = WV_SF_PCM32;
    }
    else if (CKfmt->wBitsPerSample > 0 && CKfmt->wBitsPerSample <= 8) {
      AFr->DFormat.Format = FD_UINT8;
      AFr->DFormat.ScaleF = WV_SF_PCM8;
    }
    else {
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_UnsDSize,
	      (int) CKfmt->wBitsPerSample);
      longjmp (AFR_JMPENV, 1);
    }
    break;
  case WAVE_FORMAT_MULAW:
    if (CKfmt->wBitsPerSample != 8)
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadMulaw,
	      (int) CKfmt->wBitsPerSample);
    AFr->DFormat.Format = FD_MULAW8;
    AFr->DFormat.ScaleF = WV_SF_MULAW;
    break;
  case WAVE_FORMAT_ALAW:
    if (CKfmt->wBitsPerSample != 8)
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadAlaw,
	      (int) CKfmt->wBitsPerSample);
    AFr->DFormat.Format = FD_ALAW8;
    AFr->DFormat.ScaleF = WV_SF_ALAW;
    break;
  case WAVE_FORMAT_IEEE_FLOAT:
    if (CKfmt->wBitsPerSample == 8 * FDL_FLOAT32) {
      AFr->DFormat.Format = FD_FLOAT32;
      AFr->DFormat.ScaleF = WV_SF_FLOAT32;
    }
    else if (CKfmt->wBitsPerSample == 8 * FDL_FLOAT64) {
      AFr->DFormat.Format = FD_FLOAT64;
      AFr->DFormat.ScaleF = WV_SF_FLOAT64;
    }
    else
      UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_BadFloat,
	      (int) CKfmt->wBitsPerSample);
    if (! UTcheckIEEE ())
      UTwarn ("AFrdWVhead - %s", AFM_WV_NoIEEE);
    break;

  /* Unsupported data formats */
  default:
    AF_UnsFormat ((int) CKfmt->wFormatTag);
    longjmp (AFR_JMPENV, 1);
  }

  /* Consistency check */
  if (CKfmt->nAvgBytesPerSec != CKfmt->nChannels * CKfmt->nSamplesPerSec *
      ICEILV (CKfmt->wBitsPerSample, 8))
    UTwarn ("AFrdWVhead - %s", AFM_WV_MisNBytes);

  AFr->DFormat.Swapb = DS_EL;

  AFr->Sfreq = (double) CKfmt->nSamplesPerSec;
  AFr->NData.Nchan = (long int) CKfmt->nChannels;

  return;
}

/* Read the fact chunk */

static int
AF_rdFACT (FILE *fp, struct WV_CKfact *CKfact)

{
  int offs;

  offs = RHEAD_V (fp, CKfact->ckSize, DS_EL);
  if (CKfact->ckSize < 0)
    UTwarn ("AFrdWVhead - %s", AFM_WV_BadFACT);
  else if (CKfact->ckSize >= 4)
    offs += RHEAD_V (fp, CKfact->dwSampleLength, DS_EL);
  offs += RSKIP (fp, RNDUPV (CKfact->ckSize + 4, ALIGN) - offs);

  return offs;
}

/* Unsupported data format */

static void
AF_UnsFormat (int FormatTag)

{
  switch (FormatTag) {
  case WAVE_FORMAT_ADPCM:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "G.721 ADPCM");
    break;
  case WAVE_FORMAT_MPEGLAYER3:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "MPEG-1 Layer 3");
    break;
  case WAVE_FORMAT_MSG723:
    UTwarn ("AFrdWVhead - %s: \"%s\"", AFM_WV_UnsData, "G.723.1 coding");
    break;
  default:
    UTwarn ("AFrdWVhead - %s: \"%d\"", AFM_WV_UnsData, FormatTag);
    break;
  }
}

/* Read a DISP chunk */


static int
AF_rdDISP_text (FILE *fp, int Size, struct AF_info *Hinfo)

{
  int offs;
  uint4_t DISP_ID;

  offs = RHEAD_V (fp, DISP_ID, DS_EL);
  if (DISP_ID == CF_TEXT)
    offs += AFrdHtext (fp, Size - offs, "display_name: ", Hinfo, ALIGN);

  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}     

/* Read the LIST-INFO records from the header */


struct WV_LI{
  char ckid[4];
  char *key;
};
static const struct WV_LI IID[] = {
  {"IARL", "archival_location: "},
  {"IART", "artist: "},
  {"ICMS", "commissioned: "},
  {"ICMT", "comments: "},
  {"ICOP", "copyright: "},
  {"ICRD", "creation_date: "},
  {"IENG", "engineer: "},
  {"IGNR", "genre: "},
  {"IKEY", "keywords: "},
  {"IMED", "medium: "},
  {"INAM", "name: "},
  {"IPRD", "product: "},
  {"ISBJ", "subject: "},
  {"ISFT", "software: "},
  {"ISRC", "source: "},
  {"ISRF", "source_form: "},
  {"ITCH", "technician: "}
};

#define NIID	NELEM (IID)


static int
AF_rdLIST_INFO (FILE *fp, int Size, struct AF_info *Hinfo)

{
  int i, offs;
  char ID[4], key[7];
  struct WV_CKpreamb CkHead;

  offs = RHEAD_S (fp, ID);
  if (SAME_CSTR (ID, "INFO")) {

    while (offs < Size) {
      offs += RHEAD_S (fp, CkHead.ckID);
      offs += RHEAD_V (fp, CkHead.ckSize, DS_EL);

      /* Look for standard INFO ID values */
      for (i = 0; i < (int) NIID; ++i) {
	if (SAME_CSTR (CkHead.ckID, IID[i].ckid)) {
	  offs += AFrdHtext (fp, (int) CkHead.ckSize, IID[i].key,
			     Hinfo, ALIGN);
	  break;
	}
      }

      /* No match, use the INFO ID field as the information record keyword */
      if (i == NIID) {
	strncpy (key, CkHead.ckID, 4);
	strcpy (&key[4], ": ");
	offs += AFrdHtext (fp, (int) CkHead.ckSize, key, Hinfo, ALIGN);
      }
    }
  }
  offs += RSKIP (fp, RNDUPV (Size, ALIGN) - offs);

  return offs;
}
